
/****************************************************************************
 *
 *	$RCSfile: array.hpp $
 *	$Author: tim $
 *	$Date: 1998/01/21 23:12:53 $
 *	$Revision: 1.1 $
 *
 *	Originator:
 *		Tim Foden
 *
 *	Description:
 *		TBA
 *
 ****************************************************************************/

#ifndef	_ARRAY_HPP_
#define	_ARRAY_HPP_

/*--------------------------------------------------------------------------*/

#include <memory.h>

//#include "ChessTyp.HPP"
#include "Types.hpp"
//#include "LinkList.HPP"	// for definition of INDEX
#define	INDEX	int

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
class	Array
{
public:
	enum				{ maxSize = MAX_SIZE };

protected:
	int					m_size;
	ITEM_TYPE			m_items[MAX_SIZE];

public:
	Array();

	BOOL				IsEmpty() const;
	int					GetSize() const;
	void				SetSize( int size );
	BOOL				AddTail( const ITEM_TYPE& item );
	void				AddTail( const Array<ITEM_TYPE, MAX_SIZE>& items );
	void				RemoveTail();
	void				RemoveAll();
	ITEM_TYPE&			operator[]( int index );
	const ITEM_TYPE&	operator[]( int index ) const;

	BOOL				BadPosition( INDEX pos ) const;
	INDEX				GetHeadPosition() const;
	ITEM_TYPE&			GetAt( INDEX pos );
	ITEM_TYPE&			GetNext( INDEX& pos );
	INDEX				GetNextPosition( INDEX pos ) const;
	ITEM_TYPE&			GetHead();
	const ITEM_TYPE&	GetHead() const;
	ITEM_TYPE&			GetTail();
	const ITEM_TYPE&	GetTail() const;
};

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
Array<ITEM_TYPE, MAX_SIZE>::Array()
{
	m_size = 0;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
BOOL		Array<ITEM_TYPE, MAX_SIZE>::IsEmpty() const
{
	return m_size == 0;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
int			Array<ITEM_TYPE, MAX_SIZE>::GetSize() const
{
	return m_size;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
void		Array<ITEM_TYPE, MAX_SIZE>::SetSize( int size )
{
	if( size >= 0 && size <= MAX_SIZE )
		m_size = size;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
BOOL		Array<ITEM_TYPE, MAX_SIZE>::AddTail( const ITEM_TYPE& item )
{
	DBG_ASSERT( m_size < MAX_SIZE );

	if( m_size >= MAX_SIZE )
		return FALSE;

	m_items[m_size++] = item;
	return TRUE;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
void		Array<ITEM_TYPE, MAX_SIZE>::AddTail( const Array<ITEM_TYPE, MAX_SIZE>& items )
{
	int		size = items.GetSize();
	if( size + m_size > MAX_SIZE )
	{
		DBG_ASSERT( FALSE );
		size = MAX_SIZE - m_size;
	}

	for( int i = 0; i < size; i++ )
	{
		m_items[m_size + i] = items[i];
	}

	m_size += size;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
void		Array<ITEM_TYPE, MAX_SIZE>::RemoveTail()
{
	if( m_size > 0 )
		m_size--;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
void		Array<ITEM_TYPE, MAX_SIZE>::RemoveAll()
{
	m_size = 0;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
ITEM_TYPE&		Array<ITEM_TYPE, MAX_SIZE>::operator[]( int index )
{
	DBG_ASSERT( 0 <= index && index < m_size );
	return m_items[index];
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
const ITEM_TYPE&	Array<ITEM_TYPE, MAX_SIZE>::operator[]( int index ) const
{
	DBG_ASSERT( 0 <= index && index < m_size );
	return m_items[index];
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
BOOL		Array<ITEM_TYPE, MAX_SIZE>::BadPosition( INDEX pos ) const
{
	if( pos == 0 )
		return TRUE;

	int		index = ((ITEM_TYPE*)pos) - &m_items[0];
	return index < 0 || index >= m_size;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
INDEX		Array<ITEM_TYPE, MAX_SIZE>::GetHeadPosition() const
{
	DBG_ASSERT( m_size > 0 );
	return (INDEX)&m_items[0];
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
ITEM_TYPE&		Array<ITEM_TYPE, MAX_SIZE>::GetAt( INDEX pos )
{
	return *((ITEM_TYPE*)pos);
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
ITEM_TYPE&		Array<ITEM_TYPE, MAX_SIZE>::GetNext( INDEX& pos )
{
	ITEM_TYPE*	pITEM_TYPE = ((ITEM_TYPE*)pos);
	pos = (INDEX)(pITEM_TYPE + 1);
	return *pITEM_TYPE;
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
INDEX		Array<ITEM_TYPE, MAX_SIZE>::GetNextPosition( INDEX pos ) const
{
	return (INDEX)(((ITEM_TYPE*)pos) + 1);
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
ITEM_TYPE&		Array<ITEM_TYPE, MAX_SIZE>::GetTail()
{
	DBG_ASSERT( m_size > 0 );
	return m_items[m_size - 1];
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
const ITEM_TYPE&	Array<ITEM_TYPE, MAX_SIZE>::GetTail() const
{
	DBG_ASSERT( m_size > 0 );
	return m_items[m_size - 1];
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
ITEM_TYPE&		Array<ITEM_TYPE, MAX_SIZE>::GetHead()
{
	DBG_ASSERT( m_size > 0 );
	return m_items[0];
}

/*--------------------------------------------------------------------------*/

template<class ITEM_TYPE, int MAX_SIZE>
const ITEM_TYPE&	Array<ITEM_TYPE, MAX_SIZE>::GetHead() const
{
	DBG_ASSERT( m_size > 0 );
	return m_items[0];
}

/*--------------------------------------------------------------------------*/

#endif	// !_ARRAY_HPP_
