
/****************************************************************************
 *
 *	$RCSfile: main.cpp $
 *	$Author: tim $
 *	$Date: 1998/01/21 23:12:23 $
 *	$Revision: 1.1 $
 *
 *	Originator:
 *		Tim Foden
 *
 *	Description:
 *		TBA
 *
 ****************************************************************************/
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

#ifdef  WIN32
#include <winsock.h>
#include "wsockerr.h"
#else
//#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/protocols.h>
#include <arpa/inet.h>
#endif

#include "Str.hpp"
#include "Array.hpp"

#ifndef WIN32
#define SOCKET          int
#define INVALID_SOCKET  (-1)
#define IPPROTO_UDP     (IP_UDP)
#define	INADDR_NONE		(0)
#define closesocket(s)  close(s)
#endif

#define	HTTP_PORT		(80)

/****************************************************************************/
/*
 *	global variables
 */
/*--------------------------------------------------------------------------*/

int		verboseLevel = 3;
BOOL	socketsOpen = FALSE;

/****************************************************************************/
/*
 *	functions
 */
/*--------------------------------------------------------------------------*/
/*
 *
 */

void	Log( int level, const char* pFormat, ... )
{
	if( level > verboseLevel )
		return;

	va_list	args;

	va_start( args, pFormat );
	vfprintf( stderr, pFormat, args );
	va_end( args );
}

/*--------------------------------------------------------------------------*/
/*
 *
 */

void	Usage()
{
	Log( 0, "Usage: GetHREF [<options>] <local-filename> "
			"<server-address> [<path-on-server>]\n" );
	Log( 0, "Options:\n" );
	Log( 0, "  -vn    n = level of verbosity: 0..9 -> 0=none, 9=full, default=2\n" );
	Log( 0, "  -rn-m  Retreive from byte n to byte m inclusive\n" );
	Log( 0, "  -h     Request headers only\n" );
	Log( 0, "\n" );
	Log( 0, "E.G. To retreive //www2.prestel.co.uk/diamond/foo.zip " );
	Log( 0, "you would say:\n" );
	Log( 0, "C:\\> gethref foo.zip www2.prestel.co.uk " );
	Log( 0, "/diamond/foo.zip\n" );
}

/*--------------------------------------------------------------------------*/
/*
 *
 */

String	GetErrorString()
{
#ifdef  WIN32
	return GetWSAErrorString(WSAGetLastError());
#else
	return strerr(errno);
#endif
}

/*--------------------------------------------------------------------------*/
/*
 *
 */

void    PrintError( const char* pErrStr )
{
	Log( 1, "%s: %s", pErrStr, (const char*)GetErrorString() );
}

/*--------------------------------------------------------------------------*/
/*
 *
 */

int     StartSockets()
{
#ifdef WIN32
	WSAData data;
	if( WSAStartup(MAKEWORD(1,1), &data) != 0 )
	{
		Log( 1, "Error initialising windows sockets\n" );
		return 1;
	}
#endif
	socketsOpen = TRUE;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*
 *
 */

void    EndSockets()
{
	if( !socketsOpen )
		return;

#ifdef  WIN32
	WSACleanup();
#endif

	socketsOpen = FALSE;
}

/*--------------------------------------------------------------------------*/
/*
 *
 */

void	CleanUp()
{
	Log( 9, "Cleaning up...\n" );

	EndSockets();
}

/*--------------------------------------------------------------------------*/
/*
 *
 */

void    Terminate( int sig )
{
	Log( 1, "INTERRUPT\n" );
	CleanUp();
	exit( 99 );
}

/*--------------------------------------------------------------------------*/

class	StringArray
:	Array<String, 25>
{
};

/*--------------------------------------------------------------------------*/

struct HttpRequest
{
	BOOL			m_headersOnly;		// TRUE == request headers only
	sockaddr_in		m_serverAddress;	// address of server
	String			m_pathToFile;		// name of URL to retrieve
	FILE*			m_pWriteToFile;		// writes to this file
	int				m_rangeStart;		// -1 == no range
	int				m_rangeEnd;
};

struct HttpReply
{
	int				m_errorStatus;		// 0 == OK, else error
	String			m_errorString;
	StringArray		m_headerRecords;	// header records received
	int				m_httpStatus;		// e.g. 200 == OK
	int				m_acceptRanges;		// 0 if no, else range unit value
	int				m_rangeStart;		// -1 == no range
	int				m_rangeEnd;
	int				m_nHeaderReceived;	// n bytes received in header
	int				m_nBodyReceived;	// n bytes received in body
	int				m_contentLength;	// -1 == not known
};

/*--------------------------------------------------------------------------*/
/*
 *
 */

BOOL	GetHRef( HttpRequest& request, HttpReply& reply )
{
	reply.m_errorStatus = 0;
	reply.m_httpStatus = -1;	// unknown
//	reply.m_headerRecords.RemoveAll();
	reply.m_acceptRanges = FALSE;
	reply.m_rangeStart = -1;
	reply.m_rangeEnd = -1;
	reply.m_nHeaderReceived = 0;
	reply.m_nBodyReceived = 0;
	reply.m_contentLength = -1;

	SOCKET	s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP );
	if( s == INVALID_SOCKET )
	{
		reply.m_errorStatus = 1;
		reply.m_errorString.Format(
			"socket: %s\n",
			(const char*)GetErrorString() );
		return FALSE;
	}

	Log( 9,	"Connecting to server, address=%s, port=%d...\n",
			inet_ntoa(request.m_serverAddress.sin_addr),
			ntohs(request.m_serverAddress.sin_port) );

	if( connect(s, (sockaddr*)&request.m_serverAddress,
					sizeof(request.m_serverAddress)) == SOCKET_ERROR )
	{
		reply.m_errorStatus = 2;
		reply.m_errorString.Format(
			"connect(%s:%d): %s",
			inet_ntoa(request.m_serverAddress.sin_addr),
			ntohs(request.m_serverAddress.sin_port),
			(const char*)GetErrorString() );
		return FALSE;
	}

	char	req[1000];
	strcpy( req, request.m_headersOnly ? "HEAD " : "GET " );
	strcat( req, (const char*)request.m_pathToFile );
	strcat( req, " HTTP/1.0\r\n" );
	Log( 9, "Sending request: %s", req );

	if( request.m_rangeStart != -1 )
	{
		char	rangeReq[100];
		if( request.m_rangeEnd != -1 )
			sprintf( rangeReq, "Range: bytes=%d-%d\r\n", request.m_rangeStart,
														 request.m_rangeEnd );
		else
			sprintf( rangeReq, "Range: bytes=%d-\r\n", request.m_rangeStart );
		strcat( req, rangeReq );
		Log( 9, "                 %s", rangeReq );
	}

	strcat( req, "\r\n" );

	int		nSent = send(s, req, strlen(req), 0);
	if( nSent == SOCKET_ERROR )
	{
		reply.m_errorStatus = 3;
		reply.m_errorString.Format(
			"send: %s\n",
			(const char*)GetErrorString() );
		return FALSE;
	}

	Log( 9, "%d bytes sent\n", nSent );
	Log( 8, "Waiting for reply...\n" );

	char	buf[16384];
	int		nReceived;
	char	headerLine[1000] = "";
	BOOL	firstHeader = TRUE;
	BOOL	inHeader = TRUE;
	int		contentLength = -1;
	while( (nReceived = recv(s, buf, sizeof(buf) - 1, 0)) > 0 )
	{
		buf[nReceived] = '\0';
		char*	pBuf = buf;

		if( inHeader )
		{
			char*	pEol;
			while( (pEol = strchr(pBuf, '\n')) != NULL )
			{
				int		size = (pEol + 1) - pBuf;
				nReceived -= size;
				reply.m_nHeaderReceived += size;
				*pEol = '\0';
				strcat( headerLine, pBuf );
				if( headerLine[strlen(headerLine)-1] == '\r' )
					headerLine[strlen(headerLine)-1] = '\0';
				pBuf = pEol + 1;
				if( headerLine[0] == '\0' )
				{
					inHeader = FALSE;
					Log( 9, "Receiving body...\n" );
					break;
				}

				if( firstHeader )
				{
					// this is the status line from the server
					char	version[100];
					int		status;
					if( sscanf(headerLine, "%s %d", version, &status) == 2 )
						reply.m_httpStatus = status;

					Log( 4, "** Status: %s\n", headerLine );
					firstHeader = FALSE;
				}
				else
				{
					// this is a general header line from the server
					char*	pDelim = strchr(headerLine, ':');
					if( pDelim != 0 )
					{
						pDelim++;
						while( *pDelim && *pDelim <= ' ' )
							*pDelim++;

						if( strnicmp(headerLine, "content-length", 14) == 0 )
						{
							int		value;
							if( sscanf(pDelim, "%d", &value) == 1 )
							{
								contentLength = value;
								reply.m_contentLength = value;
							}
						}
						else if( strnicmp(headerLine, "accept-ranges", 13) == 0 )
						{
							int		value;
							if( sscanf(pDelim, "%d", &value) == 1 )
								reply.m_acceptRanges = value;
							else if( strnicmp(pDelim, "bytes", 5 ) == 0 )
								reply.m_acceptRanges = 1;
						}
					}

					Log( 5, "** Header: %s\n", headerLine );
				}

//				request.m_headerRecords.AddTail( headerLine );
				headerLine[0] = '\0';
			}

			if( inHeader )
			{
				strcat( headerLine, pBuf );
				reply.m_nHeaderReceived += nReceived;
				nReceived = 0;
			}
		}

		if( !inHeader && nReceived > 0 )
		{
			fwrite( pBuf, nReceived, 1, request.m_pWriteToFile );
			reply.m_nBodyReceived += nReceived;
			Log( 3, "\rReceived %d/%d (%1.1f%%)...",
					reply.m_nBodyReceived, contentLength,
					contentLength > 0 ?
						100.0 * reply.m_nBodyReceived / contentLength : 0 );
			fflush( stdout );
		}
	}

	Log( 3, "\n" );

	if( nReceived == SOCKET_ERROR )
	{
		reply.m_errorStatus = 4;
		reply.m_errorString.Format(
			"recv: %s\n",
			(const char*)GetErrorString() );
		return FALSE;
	}

//	Log( 9, "Total bytes received: %d\n", reply.m_nReceived );

	reply.m_errorStatus = 0;
	return TRUE;
}

/*--------------------------------------------------------------------------*/
/*
 *
 */

int		main( int argc, char** argv )
{
	//
	//	parse input parameters
	//

	String	filename;
	String	serverAddress;
	String	URI = "/";
	int		rangeLow = -1;	// no range
	int		rangeHigh = -1;
	BOOL	headersOnly = FALSE;
	int		count = 0;

	for( int i = 1; i < argc; i++ )
	{
		if( argv[i][0] == '-' )
		{
			switch( argv[i][1] )
			{
				case 'v':
				case 'V':
					if( argv[i][2] >= '0' && argv[i][2] <= '9' )
						verboseLevel = argv[i][2] - '0';
					break;

				case 'r':
				case 'R':
					if( sscanf(argv[i] + 2, "%d-%d", &rangeLow, &rangeHigh) != 2)
					{
						rangeLow = -1;
						rangeHigh = -1;
					}
					break;

				case 'h':
				case 'H':
					headersOnly = TRUE;
					break;
			}
		}
		else
		{
			switch( count )
			{
				case 0:	filename = argv[i];			break;
				case 1:	serverAddress = argv[i];	break;
				case 2: URI = argv[i];				break;
			}

			count++;
		}
	}

	if( count < 2 || count > 3 )
	{
		Usage();
		return 1;
	}

	//
	//	set up to use sockets
	//

	signal( SIGINT, Terminate );

	int	status = StartSockets();
	if( status != 0 )
		return status;

	//
	//	set up the server address
	//

	int			portNumber = HTTP_PORT;

	sockaddr_in	addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(portNumber);
	addr.sin_addr.s_addr = inet_addr(serverAddress);
	if( addr.sin_addr.s_addr == INADDR_NONE )
	{
		// try to lookup the name...
		Log( 9, "Looking up address of host \"%s\"...\n",
				(const char*)serverAddress );
		hostent*	pHost = gethostbyname(serverAddress);
		if( pHost == 0 )
		{
			PrintError( "GetHostByName" );
			CleanUp();
			return 2;
		}

		memcpy( &addr.sin_addr.s_addr, pHost->h_addr_list[0],
				pHost->h_length );
	}

	//
	//	set up the HTTP request
	//

	HttpRequest	request;
	request.m_headersOnly = headersOnly;
	request.m_serverAddress = addr;
	request.m_pathToFile = URI;
	request.m_rangeStart = rangeLow;
	request.m_rangeEnd = rangeHigh;
//	request.m_rangeStart = 0;	//
//	request.m_rangeEnd = 99;	// 

	request.m_pWriteToFile = fopen(filename, "wb");
	if( request.m_pWriteToFile == 0 )
	{
		PrintError( "fopen" );
		CleanUp();
		return 4;
	}

	//
	//	perform the request
	//

	HttpReply	reply;
	if( !GetHRef(request, reply) )
	{
		int		rangeUnits = reply.m_acceptRanges;
		int		totalSoFar = reply.m_nBodyReceived;
		int		maxTries = 3;
		int		minDelta = 25 * 1024;	// 25K
		int		nTries = 0;

		while( nTries < maxTries && reply.m_errorStatus != 0 )
		{
			Log( 1, "%s\n", (const char*)reply.m_errorString );

			// make up a range data request or do the whole request
			// again
			if( rangeUnits != 0 )
			{
				// make up a range request
				request.m_rangeStart = totalSoFar;
				request.m_rangeEnd = -1;	// to the end
			}
			else
			{
				// do the whole request again
				rewind( request.m_pWriteToFile );
			}

			if( !GetHRef(request, reply) )
			{
				if( reply.m_nBodyReceived >= minDelta )
					nTries = 0;
				nTries++;
			}

			totalSoFar += reply.m_nBodyReceived;
		}

		CleanUp();
		return 10 + reply.m_errorStatus;
	}

	fclose( request.m_pWriteToFile );

	//
	//	dump some info from the reply
	///

	if( reply.m_httpStatus != -1 )
		Log( 5, "Reply status: %d\n", reply.m_httpStatus );

	if( reply.m_acceptRanges != 0 )
		Log( 5, "Ranges accepted: resolution=%d\n", reply.m_acceptRanges );

	if( reply.m_contentLength != -1 )
		Log( 5, "Content length: %d\n", reply.m_contentLength );
	else
		Log( 5, "Content length: Unknown\n" );

	Log( 5, "Header bytes received: %d\n", reply.m_nHeaderReceived );
	Log( 5, "Body bytes received: %d\n", reply.m_nBodyReceived );

	//
	// cleanup
	//

	CleanUp();
	return 0;
}

/*--------------------------------------------------------------------------*/
