
/****************************************************************************
 *
 *	$RCSfile: str.cpp $
 *	$Author: tim $
 *	$Date: 1998/01/21 23:12:23 $
 *	$Revision: 1.1 $
 *
 *	Originator:
 *		Tim Foden
 *
 *	Description:
 *		TBA
 *
 ****************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <memory.h>

#include "str.hpp"

char	String::m_emptyString = '\0';

void	String::InitOnce()
{
	m_caseSignificant = TRUE;
	Initialise();
}

void	String::Initialise()
{
	m_length = 0;
	m_pData = 0;
}

void	String::Free()
{
	delete [] m_pData;

	m_length = 0;
	m_pData = 0;
}

void	String::Normalise( int& index ) const
{
	if( index < 0 )
	{
		index += m_length;
		if( index < 0 ) index = 0;
	}
	else if( index >= m_length )
	{
		index = m_length - 1;
	}
}

String::String()
{
	InitOnce();
	SetLength( 0 );
}

String::String( const char* pBuffer, int length )
{
	InitOnce();
	Set( pBuffer, length );
}

String::String( const char* pString )
{
	InitOnce();
	Set( pString );
}

String::String( const String& string )
{
	InitOnce();
	Set( string );
}

String::String( char ch )
{
	InitOnce();
	Set( ch );
}

String::~String()
{
	Free();
}

void	String::Empty()
{
	SetLength( 0 );
}

void	String::SetLength( int newLength )
{
	// nothing to do?
	if( newLength == m_length )
		return;

	// deallocate the old data
	Free();

	// treat negative length strings as zero length
	if( newLength < 0 )
		newLength = 0;

	// trap silly lengths and make them sensible
	if( newLength > 65534 )
	{
//		DBG_ASSERT( FALSE );
		newLength = 65534;
	}

	// try to allocate the new memory block
	m_pData = new char[newLength + 1];
	if( m_pData == 0 )
		return;

	// update member variables
	m_length = newLength;
	m_pData[newLength] = '\0';
}

void	String::Set( const char* pBuffer, int bufferLen )
{
	if( bufferLen < 0 || pBuffer == 0 )
		bufferLen = 0;

	SetLength( bufferLen );

	if( m_pData != 0 )
		memcpy( m_pData, pBuffer, m_length );
}

void	String::Set( const char* pString )
{
	Set( pString, pString == 0 ? 0 : strlen(pString) );
}

void	String::Set( const String& string )
{
	Set( string.m_pData, string.m_length );
}

void	String::Set( char ch )
{
	SetLength( 1 );

	if( m_pData != 0 )
		m_pData[0] = ch;
}

void	String::Slice( int start, int end, String& string ) const
{
	if( m_pData == 0 )
	{
		string.Empty();
		return;
	}

	Normalise( start );
	Normalise( end );

	if( start > end )
	{
		string.Empty();
		return;
	}

	// otherwise 0 <= start <= end < m_length, so
	string.Set( m_pData + start, end - start + 1 );
}

void	String::Delete( int start, int end )
{
	if( m_pData == 0 )
		return;

	Normalise( start );
	Normalise( end );

	if( start > end ) return;

	// simply copy the excess data down to fill the gap, and modify
	// the length
	memcpy( m_pData + end + 1, m_pData + start, m_length - (end + 1) );
	m_length -= (end - start + 1);
	m_pData[m_length] = '\0';
}

void	String::Remove( int start, int end, String& string )
{
	Slice( start, end, string );
	Delete( start, end );
}

void	String::RemoveExcluding( char ch, String& string, BOOL forwards )
{
	int		position = Find(ch, forwards ? 0 : -1, forwards);

	// found ?
	if( position >= 0 )
	{
		// searching forwards?
		if( forwards )
		{
			// remove from start of string
			Remove( 0, position - 1, string );
		}
		else
		{
			// remove from end of string
			Remove( position + 1, -1, string );
		}
	}
	else
	{
		// remove the whole string
		string.Set( *this );
		Empty();
	}
}

void	String::RemoveExcluding( const String& charSet, String& string, BOOL forwards )
{
	// only forwards supported at this time
	if( !forwards )
		return;

	int		position = FindSet(charSet, forwards ? 0 : -1, forwards);

	// found ?
	if( position >= 0 )
	{
		// searching forwards?
		if( forwards )
		{
			// remove from start of string
			Remove( 0, position - 1, string );
		}
		else
		{
			// remove from end of string
			Remove( position + 1, -1, string );
		}
	}
	else
	{
		// remove the whole string
		string.Set( *this );
		Empty();
	}
}

void	String::RemoveIncluding( const String& charSet, String& string, BOOL forwards )
{
	// only forwards supported at this time
	if( !forwards )
		return;

	int		position = FindNotSet(charSet, forwards ? 0 : -1, forwards);

	// found ?
	if( position >= 0 )
	{
		// searching forwards?
		if( forwards )
		{
			// remove from start of string
			Remove( 0, position - 1, string );
		}
		else
		{
			// remove from end of string
			Remove( position + 1, -1, string );
		}
	}
	else
	{
		// remove the whole string
		string.Set( *this );
		Empty();
	}
}
/*
String	String::RemoveField( const String& delimSet, BOOL multiDelim )
{
	String	field;

	RemoveExcluding( delimSet, field );

	if( multiDelim )
		DeleteIncluding( delimSet );
	else
		Delete( 1, 2 );

	return field;
}
*/
String	String::Slice( int start, int end ) const
{
	String	string;

	Slice( start, end, string );

	return string;
}

String	String::Left( int nChars ) const
{
	return Slice(0, nChars - 1);
}

String	String::Mid( int index ) const
{
	return Slice(index, m_length - 1);
}

String	String::Mid( int index, int nChars ) const
{
	return Slice(index, index + nChars - 1);
}

String	String::Right( int nChars ) const
{
	return Slice(m_length - nChars, m_length - 1);
}

int		String::Find( char ch, int start, BOOL forwards ) const
{
	if( m_pData == 0 )
		return -1;

	Normalise( start );

	BOOL	found = FALSE;
	int		incr  = forwards ? 1 : -1;
	int		count = forwards ? m_length - start : start + 1;

	while( count-- > 0 )
	{
		if( m_pData[start] == ch )
		{
			found = TRUE;
			break;
		}
		else
		{
			start += incr;
		}
	}

	return found ? start : -1;
}

int		String::Find( const String& subString, int start ) const
{
	if( subString.IsEmpty() )
		return start;

	Normalise( start );
	int		maxLen = m_length - subString.GetSize();

	for( int i = start; i < maxLen; i++ )
	{
		if( subString[0] == m_pData[i] )
		{
			// test if a match
			BOOL	found = TRUE;
			for( int j = 1; j < subString.GetSize(); j++ )
			{
				if( subString[j] != m_pData[i + j] )
				{
					found = FALSE;
					break;
				}
			}

			if( found )
				return i;
		}
	}

	return -1;   // couldn't find substring at all
}

int		String::FindReverse( const String& /*subString*/, int /*start*/ ) const
{
	return -1;   // couldn't find substring at all
}

int   	String::FindSet(	const String& charSet, int start,
							BOOL forwards ) const
{
	// only forwards searches supported at this time
	if( !forwards || m_pData == 0 )
		return -1;

	char*	pCh = strpbrk(m_pData + start, charSet);
	if( pCh == 0 )
		return -1;

	return pCh - m_pData;
}

int   	String::FindNotSet(	const String& charSet, int start,
							BOOL forwards) const
{
	// only forwards searches supported at this time
	if( !forwards || m_pData == 0 )
		return -1;

	int		len = strcspn(m_pData + start, charSet);
	if( len == 0 )
		return -1;

	return start + len;
}

void    String::Append( const String& string )
{
	const char*	pBuffer		= string.m_pData;
	int			bufferLen	= string.m_length;
	char*   	pData;

	// check there is a string to add
	if( pBuffer == 0 || bufferLen == 0 ) return;

	// truncate if will be too long
	if( bufferLen + m_length > maxLength )
		bufferLen = maxLength - m_length;

	// allocate the larger buffer
	pData = new char[m_length + bufferLen + 1];
	if( pData == 0 )
		return;

	// copy the old data from the old to the new buffer
	memcpy( pData, m_pData, m_length );

	// append the new data
	memcpy( pData + m_length, pBuffer, bufferLen );

	// deallocate the old buffer
	delete [] m_pData;

	// adjust internals for new string
	m_length += bufferLen;
	m_pData = pData;
	m_pData[m_length] = '\0';
}

void	String::FormatV( const char* pFormat, va_list args )
{
	char	line[1000];
	vsprintf( line, pFormat, args );
	Set( line );
}

void	String::Format( const char* pFormat, ... )
{
	va_list	argl;
	va_start( argl, pFormat );
	FormatV( pFormat, argl );
	va_end( argl );
}

String&	String::operator=( const String& string )
{
	Set( string );
	return *this;
}

String&	String::operator+=( const String& string )
{
	Append( string );
	return *this;
}

String	String::operator+( const String& string )
{
	String	temp(*this);
	temp.Append( string );
	return temp;
}

String::operator const char*() const
{
	return m_pData != 0 ? m_pData : &m_emptyString;
}

char	String::operator[]( int index ) const
{
	if( index < 0 ) index += m_length;

	if( index < 0 || index >= m_length ) return '\0';

	return m_pData[index];
}

char&	String::operator[]( int index )
{
	static char	errorCh = '\0';

	if( index < 0 ) index += m_length;

	if( index < 0 || index >= m_length ) return errorCh;

	return m_pData[index];
}

BOOL	String::operator==( const String& string ) const
{
	if( m_caseSignificant )
		return strcmp(m_pData, string.m_pData) == 0;

	return stricmp(m_pData, string.m_pData) == 0;
}

BOOL	String::operator!=( const String& string ) const
{
	if( m_caseSignificant )
		return strcmp(m_pData, string.m_pData) != 0;

	return stricmp(m_pData, string.m_pData) != 0;
}

BOOL	String::operator< ( const String& string ) const
{
	if( m_caseSignificant )
		return strcmp(m_pData, string.m_pData) < 0;

	return stricmp(m_pData, string.m_pData) < 0;
}

BOOL	String::operator>=( const String& string ) const
{
	if( m_caseSignificant )
		return strcmp(m_pData, string.m_pData) >= 0;

	return stricmp(m_pData, string.m_pData) >= 0;
}

BOOL	String::operator> ( const String& string ) const
{
	if( m_caseSignificant )
		return strcmp(m_pData, string.m_pData) > 0;

	return stricmp(m_pData, string.m_pData) > 0;
}

BOOL	String::operator<=( const String& string ) const
{
	if( m_caseSignificant )
		return strcmp(m_pData, string.m_pData) <= 0;

	return stricmp(m_pData, string.m_pData) <= 0;
}

BOOL	String::operator==( const char* pString ) const
{
	return *this == String(pString);
}

BOOL	String::operator!=( const char* pString ) const
{
	return *this != String(pString);
}

BOOL	String::operator< ( const char* pString ) const
{
	return *this < String(pString);
}

BOOL	String::operator>=( const char* pString ) const
{
	return *this >= String(pString);
}

BOOL	String::operator> ( const char* pString ) const
{
	return *this > String(pString);
}

BOOL	String::operator<=( const char* pString ) const
{
	return *this <= String(pString);
}


