
/****************************************************************************
 *
 *	$RCSfile: str.hpp $
 *	$Author: tim $
 *	$Date: 1998/01/21 23:12:23 $
 *	$Revision: 1.1 $
 *
 *	Originator:
 *		Tim Foden
 *
 *	Description:
 *		TBA
 *
 ****************************************************************************/

#ifndef _STR_HPP_
#define _STR_HPP_

#include <stdarg.h>

#include "Types.hpp"

class String
{
private:
	int			m_length;
	char*		m_pData;
	BOOL		m_caseSignificant;

	static char	m_emptyString;

public:
	enum	{ maxLength = 32767 };

private:
	void		InitOnce();			// initialise once only
	void		Initialise();		// initialise member variables
	void		Free();				// deallocates storage
	void		Normalise( int& index ) const;

public:
	String();
	String( const char* pBuffer, int length );
	String( const char* pString );
	String( const String& string );
	String( char ch );
	~String();

public:
	void		Empty();			// zeros string length
	BOOL		IsEmpty() const		{ return m_length == 0; }

	int			GetLength() const	{ return m_length; }
	void		SetLength( int newLength );

	int			GetSize() const		{ return m_length; }
	void		SetSize( int newLength );

	void		SetCaseSignificant( BOOL sig )	{ m_caseSignificant = sig; }
	BOOL		GetCaseSignificant() const		{ return m_caseSignificant; }

	void		Set( const char* pBuffer, int bufferLen );
	void		Set( const char* pString );
	void		Set( const String& string );
	void		Set( char ch );

	void		Slice(  int start, int end, String& string ) const;
	void		Delete( int start, int end );
	void		Remove( int start, int end, String& string );

	void		RemoveExcluding( char ch, String& string, BOOL forwards = TRUE );
	void		RemoveExcluding( const String& charSet, String& string, BOOL forwards = TRUE );
	void		RemoveIncluding( const String& charSet, String& string, BOOL forwards = TRUE );

	String		Slice(  int start, int end ) const;
	String		Left(   int nChars ) const;
	String		Mid(    int index ) const;
	String		Mid(    int index, int nChars ) const;
	String		Right(  int nChars ) const;

	int   		Find(   char charToFind, int start = 0,
						BOOL forwards = TRUE ) const;
	int   		Find( const String& subString, int start = 0 ) const;
	int   		FindReverse( const String& subString, int start = 0 ) const;
	int   		FindSet(   const String& charSet, int start = 0,
						BOOL forwards = TRUE ) const;
	int   		FindNotSet(   const String& charSet, int start = 0,
						BOOL forwards = TRUE ) const;

	void        Append( const String& string );

	void		Format( const char* pFormat, ... );
	void		FormatV( const char* pFormat, va_list args );

	String&		operator= ( const String& string );
	String&		operator+=( const String& string );
	String		operator+ ( const String& string );

				operator const char* () const;
	char  		operator[]( int index ) const;
	char&  		operator[]( int index );

	BOOL		operator==( const String& string ) const;
	BOOL		operator!=( const String& string ) const;
	BOOL		operator< ( const String& string ) const;
	BOOL		operator>=( const String& string ) const;
	BOOL		operator> ( const String& string ) const;
	BOOL		operator<=( const String& string ) const;

	BOOL		operator==( const char* pString ) const;
	BOOL		operator!=( const char* pString ) const;
	BOOL		operator< ( const char* pString ) const;
	BOOL		operator>=( const char* pString ) const;
	BOOL		operator> ( const char* pString ) const;
	BOOL		operator<=( const char* pString ) const;
};

#endif  // _STR_HPP_

